@@ -11,6 +11,7 @@
 
 #include "config.h"
 #include <stdarg.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -203,7 +204,7 @@ sbuf_catb(safebuf_t *s1, const safebuf_t *s2) {
     size_t res;
     assert(SAFEBUF_CHECK(s1));
     assert(SAFEBUF_CHECK(s2));
-    s2len = sbuf_len(s2);
+    s2len = (size_t)sbuf_len(s2);
     sbuf_grow(s1, 1 + s1->cl + s2len);
     res = strlcat(s1->buf + s1->cl, s2->buf, s1->len - s1->cl);
     assert( res < s1->len );
@@ -212,10 +213,10 @@ sbuf_catb(safebuf_t *s1, const safebuf_t *s2) {
 }
 
 /* Return length of string in sbuf */
-size_t
+int
 sbuf_len(const safebuf_t *sb) {
     assert(SAFEBUF_CHECK(sb));
-    return sb->cl;
+    return (int)sb->cl;
 }
 
 /* Return C string in an sbuf */
@@ -415,7 +416,7 @@ count_udtypes(int ncid) {
 	/* Get number of types in this group */
 	NC_CHECK( nc_inq_typeids(ncid, &ntypes, NULL) ) ;
 	NC_CHECK( nc_inq_grps(ncid, &numgrps, NULL) ) ;
-	ncids = (int *) emalloc(sizeof(int) * (numgrps + 1));
+	ncids = (int *) emalloc(sizeof(int) * (size_t)(numgrps + 1));
 	NC_CHECK( nc_inq_grps(ncid, NULL, ncids) ) ;
 	/* Add number of types in each subgroup, if any */
 	for (i=0; i < numgrps; i++) {
@@ -536,7 +537,7 @@ ncfloat_val_equals(const nctype_t *this,
 		   const void *v1p, const void *v2p) {
     float v1 = *(float* )v1p;
     float v2 = *(float* )v2p;
-    if((v1 > 0.0f) != (v2 > 0.0f))	/* avoid overflow */
+    if((v1 > 0.0F) != (v2 > 0.0F))	/* avoid overflow */
 	return false;
     if(isfinite(v1) && isfinite(v2))
 	return (absval(v1 - v2) <= absval(float_eps * v2)) ;
@@ -656,8 +657,8 @@ ncvlen_val_equals(const nctype_t *this,
 bool_t
 nccomp_val_equals(const nctype_t *this,
 		  const void *v1p, const void *v2p) {
-    int nfields = this->nfields;
-    int fidx;			/* field id */
+    size_t nfields = this->nfields;
+    size_t fidx;			/* field id */
 
     for (fidx = 0; fidx < nfields; fidx++) {
 	size_t offset = this->offsets[fidx];
@@ -696,8 +697,11 @@ ncbyte_typ_tostring(const nctype_t *typ, safebuf_t *sfbf, const void *valp) {
 int
 ncchar_typ_tostring(const nctype_t *typ, safebuf_t *sfbf, const void *valp) {
     char sout[PRIM_LEN];
+    char cstr[2];
     int res;
-    res = snprintf(sout, PRIM_LEN, typ->fmt, *(char *)valp);
+cstr[0] = *(char*)valp;
+cstr[1] = '\0';
+    res = snprintf(sout, PRIM_LEN, typ->fmt, cstr);
     assert(res < PRIM_LEN);
     sbuf_cpy(sfbf, sout);
     return sbuf_len(sfbf);
@@ -858,7 +862,7 @@ int ncstring_typ_tostring(const nctype_t *typ, safebuf_t *sfbf, const void *valp
         sp = sout;
         *sp++ = '"' ;
         while(*cp) {
-            switch (uc = *cp++ & 0377) {
+            switch (uc = (unsigned char)*cp++ & 0377) {
             case '\b':
                 *sp++ = '\\';
                 *sp++ = 'b' ;
@@ -901,7 +905,7 @@ int ncstring_typ_tostring(const nctype_t *typ, safebuf_t *sfbf, const void *valp
                     sp += 4;
                 }
                 else
-                    *sp++ = uc;
+                    *sp++ = (char)uc;
                 break;
             }
         }
@@ -965,7 +969,7 @@ ncopaque_val_as_hex(size_t size, char *sout, const void *valp) {
     char *sp = sout;
     int i;
     char *prefix = "0X";
-    int prelen = strlen(prefix);
+    size_t prelen = strlen(prefix);
 
     snprintf(sp, prelen + 1, "%s", prefix);
     sp += prelen;
@@ -976,7 +980,7 @@ ncopaque_val_as_hex(size_t size, char *sout, const void *valp) {
 	sp += 2;
     }
     *sp = '\0';
-    return 2*size + prelen;
+    return (int)(2*size + prelen);
 }
 
 /* Convert an opaque value to a string, represented as hexadecimal
@@ -1031,7 +1035,8 @@ chars_tostring(
 {
     long iel;
     const char *sp;
-    char *sout = (char *)emalloc(4*len + 5); /* max len of string */
+    size_t sout_size = 4*len + 5; /* max len of string */
+    char *sout = (char *)emalloc(sout_size);
     char *cp = sout;
     *cp++ = '"';
 
@@ -1041,7 +1046,7 @@ chars_tostring(
 	len--;
     for (iel = 0; iel < len; iel++) {
 	unsigned char uc;
-	switch (uc = *vals++ & 0377) {
+	switch (uc = (unsigned char)(*vals++ & 0377)) {
 	case '\b':
 	case '\f':
 	case '\n':
@@ -1058,7 +1063,8 @@ chars_tostring(
 	    if (isprint(uc))
 		*cp++ = *(char *)&uc; /* just copy, even if char is signed */
 	    else {
-		sprintf(cp,"\\%.3o",uc);
+		size_t remaining = sout_size - (size_t)(cp - sout);
+		snprintf(cp,remaining,"\\%.3o",uc);
 		cp += 4;
 	    }
 	    break;
@@ -1076,8 +1082,8 @@ chars_tostring(
    each member field */
 int
 nccomp_typ_tostring(const nctype_t *tinfo, safebuf_t *sfbf, const void *valp) {
-    int nfields = tinfo->nfields;
-    int fidx;			/* field id */
+    size_t nfields = tinfo->nfields;
+    size_t fidx;			/* field id */
     safebuf_t* sout2 = sbuf_new();
 
     sbuf_cpy(sfbf, "{");
@@ -1100,11 +1106,9 @@ nccomp_typ_tostring(const nctype_t *tinfo, safebuf_t *sfbf, const void *valp) {
 	    sbuf_cpy(sout2, "{");
 	    if(finfo->tid == NC_CHAR) { /* aggregate char rows into strings */
 		int rank = tinfo->ranks[fidx];
-		size_t nstrings;
-		size_t slen;
 		int j;
-		slen = tinfo->sides[fidx][rank-1];
-		nstrings = 1;	/* product of all but last array dimension */
+		size_t slen = (size_t)tinfo->sides[fidx][rank-1];
+		int nstrings = 1;	/* product of all but last array dimension */
 		for(j=0; j < rank-1; j++) {
 		    nstrings *= tinfo->sides[fidx][j];
 		}
@@ -1131,7 +1135,7 @@ nccomp_typ_tostring(const nctype_t *tinfo, safebuf_t *sfbf, const void *valp) {
 	    sbuf_free(sout3);
 	}
 	sbuf_catb(sfbf, sout2);
-	if(fidx < nfields - 1) {
+	if((fidx + 1) < nfields) {
 	    sbuf_cat(sfbf, ", ");
 	}
     }
@@ -1242,10 +1246,10 @@ double to_double(const ncvar_t *varp, const void *valp) {
 	dd = *(unsigned int *)valp;
 	break;
     case NC_INT64:
-	dd = *(long long *)valp;
+	dd = (double)*(long long *)valp;
 	break;
     case NC_UINT64:
-	dd = *(unsigned long long *)valp;
+	dd = (double)*(unsigned long long *)valp;
 	break;
     default:
 	error("to_double: type not numeric primitive");
@@ -1261,13 +1265,13 @@ nctime_val_tostring(const ncvar_t *varp, safebuf_t *sfbf, const void *valp) {
     if(isfinite(vv)) {
 	int oldopts = 0;
 	int newopts = 0;
-	int res;
+	size_t res;
 	sout[0]='"';
 	/* Make nctime dump error messages */
 	oldopts = cdSetErrOpts(0);
 	newopts = oldopts | CU_VERBOSE;
 	cdSetErrOpts(newopts);
-	cdRel2Iso(varp->timeinfo->calendar, varp->timeinfo->units, separator, vv, &sout[1]);
+	cdRel2Iso(varp->timeinfo->calendar, varp->timeinfo->units, separator, vv, &sout[1], sizeof(sout) - 1);
 	cdSetErrOpts(oldopts);
 	res = strlen(sout);
 	sout[res++] = '"';
@@ -1436,7 +1440,6 @@ set_tostring_func(ncvar_t *varp) {
 	      varp->tinfo->class);
     }
 #endif /* USE_NETCDF4 */
-    return;
 }
 
 
@@ -1517,7 +1520,7 @@ init_types(int ncid) {
     if (max_type == 0) {	/* if called for first time */
 	int maxtype = max_typeid(ncid);
 	int i;
-	nctypes = (nctype_t **) emalloc((maxtype + 2) * sizeof(nctype_t *));
+	nctypes = (nctype_t **) emalloc((size_t)(maxtype + 2) * sizeof(nctype_t *));
 	for(i=0; i < maxtype+1; i++)
 	    nctypes[i] = NULL;	/* so can later skip over unused type slots */
 	init_prim_types(ncid);
@@ -1529,7 +1532,7 @@ init_types(int ncid) {
    if (ntypes)
    {
       int t;
-      int *typeids = emalloc((ntypes + 1) * sizeof(int));
+      int *typeids = emalloc((size_t)(ntypes + 1) * sizeof(int));
       NC_CHECK( nc_inq_typeids(ncid, NULL, typeids) );
       for (t = 0; t < ntypes; t++) {
 	  nctype_t *tinfo;	/* details about the type */
@@ -1584,14 +1587,14 @@ init_types(int ncid) {
 		  NC_CHECK( nc_inq_compound_field(ncid, tinfo->tid, fidx, NULL,
 						  &offset, &ftype, &rank,
 						  sides) );
-		  if(rank > 0) sides = (int *) emalloc(rank * sizeof(int));
+		  if(rank > 0) sides = (int *) emalloc((size_t)rank * sizeof(int));
 		  NC_CHECK( nc_inq_compound_field(ncid, tinfo->tid, fidx, NULL,
 						  NULL, NULL, NULL, sides) );
 		  tinfo->fids[fidx] = ftype;
 		  tinfo->offsets[fidx] = offset;
 		  tinfo->ranks[fidx] = rank;
 		  if (rank > 0)
-		      tinfo->sides[fidx] = (int *) emalloc(rank * sizeof(int));
+		      tinfo->sides[fidx] = (int *) emalloc((size_t)rank * sizeof(int));
 		  tinfo->nvals[fidx] = 1;
 		  for(i = 0; i < rank; i++) {
 		      tinfo->sides[fidx][i] = sides[i];
@@ -1626,7 +1629,7 @@ init_types(int ncid) {
       /* See how many groups there are. */
       NC_CHECK( nc_inq_grps(ncid, &numgrps, NULL) );
       if (numgrps > 0) {
-	  ncids = (int *) emalloc(numgrps * sizeof(int));
+	  ncids = (int *) emalloc((size_t)numgrps * sizeof(int));
 	  /* Get the list of group ids. */
 	  NC_CHECK( nc_inq_grps(ncid, NULL, ncids) );
 	  /* Call this function for each group. */
@@ -1667,10 +1670,10 @@ iscoordvar(int ncid, int varid)
 #endif
 	if (dims)
 	    free(dims);
-	dims = (ncdim_t *) emalloc((ndims + 1) * sizeof(ncdim_t));
+	dims = (ncdim_t *) emalloc((size_t)(ndims + 1) * sizeof(ncdim_t));
 	if (dimids)
 	    free(dimids);
-	dimids = (int *) emalloc((ndims + 1) * sizeof(int));
+	dimids = (int *) emalloc((size_t)(ndims + 1) * sizeof(int));
 #ifdef USE_NETCDF4
 	NC_CHECK( nc_inq_dimids(ncid, &ndims1, dimids, include_parents ) );
 #else
@@ -1764,7 +1767,7 @@ print_type_name(int locid, int typeid) {
 	    curlocid = parent_groupid;
 	    NC_CHECK( nc_inq_typeids(curlocid, &ntypes, NULL) );
 	    if(ntypes > 0) {
-		int *typeids = (int *) emalloc((ntypes + 1) * sizeof(int));
+		int *typeids = (int *) emalloc((size_t)(ntypes + 1) * sizeof(int));
 		int i;
 		NC_CHECK( nc_inq_typeids(curlocid, &ntypes, typeids) );
 		for(i = 0; i < ntypes; i++) {
@@ -1801,9 +1804,8 @@ print_type_name(int locid, int typeid) {
 static int
 init_is_unlim(int ncid, int **is_unlim_p)
 {
-    int num_grps;	 /* total number of groups */
-    int num_dims = 0;    /* total number of dimensions in all groups */
-    int num_undims = 0;  /* total number of unlimited dimensions in all groups */
+    size_t num_grps;	 /* total number of groups */
+    int max_dimid = -1;    /* maximum dimid across whole dataset */
     int *grpids = NULL;	 /* temporary list of all grpids */
     int igrp;
     int grpid;
@@ -1822,29 +1824,40 @@ init_is_unlim(int ncid, int **is_unlim_p)
     NC_CHECK( nc_inq_grps_full(ncid, &num_grps, grpids) );
 #define DONT_INCLUDE_PARENTS 0
     /* Get all dimensions in groups and info about which ones are unlimited */
+    /* Warning: we cannot assume that the dimension ids are packed */
+    /* Find maximum dimension id */
+    max_dimid = -1;
     for(igrp = 0; igrp < num_grps; igrp++) {
-	int ndims;
+	int i,ndims;
+	int* dimids = NULL;
 	grpid = grpids[igrp];
 	NC_CHECK( nc_inq_dimids(grpid, &ndims, NULL, DONT_INCLUDE_PARENTS) );
-	num_dims += ndims;
+	dimids = (int*)emalloc((size_t)ndims*sizeof(int));
+	NC_CHECK( nc_inq_dimids(grpid, &ndims, dimids, DONT_INCLUDE_PARENTS) );
+	for(i=0;i<ndims;i++) {if(dimids[i] > max_dimid) max_dimid = dimids[i];}
+	free(dimids);
     }
-    *is_unlim_p = emalloc((num_dims + 1) * sizeof(int));
+    assert(max_dimid >= 0);
+    *is_unlim_p = emalloc((size_t)(max_dimid + 1 + 1) * sizeof(int));
     for(igrp = 0; igrp < num_grps; igrp++) {
 	int ndims, idim, *dimids, nundims;
 	grpid = grpids[igrp];
 	NC_CHECK( nc_inq_dimids(grpid, &ndims, NULL, DONT_INCLUDE_PARENTS) );
-	dimids = emalloc((ndims + 1) * sizeof(int));
+	dimids = emalloc((size_t)(ndims + 1) * sizeof(int));
 	NC_CHECK( nc_inq_dimids(grpid, &ndims, dimids, DONT_INCLUDE_PARENTS) );
 	/* mark all dims in this group as fixed-size */
 	for(idim = 0; idim < ndims; idim++) {
-	    (*is_unlim_p)[dimids[idim]] = 0;
+	    int* isunlim = *is_unlim_p;
+	    int did = dimids[idim];
+	    isunlim[did] = 0;
 	}
 	NC_CHECK( nc_inq_unlimdims(grpid, &nundims, dimids) );
 	assert(nundims <= ndims);
 	/* mark the subset of dims in this group that are unlimited */
 	for(idim = 0; idim < nundims; idim++) {
-	    (*is_unlim_p)[dimids[idim]] = 1;
-	    num_undims++;
+	    int* isunlim = *is_unlim_p;
+	    int did = dimids[idim];
+	    isunlim[did] = 1;
 	}
 	if(dimids)
 	    free(dimids);
